module libdsl.common;
import std.algorithm;
import std.conv;
import std.traits;
import std.typecons;
import std.typetuple;
import libdsl.package_common;
import libdsl.parser;

class Token{
	bool isEmpty = true;
	size_t line, num;
	public final string at(){
		import std.conv;
		return "("~line.to!string~") "~num.to!string;
	}

	private static class TokenImpl(T) : Token{
		public T value;
		private TypeInfo typeinfo = typeid(T);
		private enum needHash = !is(typeof(value.toHash()));
		static if(needHash) hash_t hash;
		private auto getHash(){
			if(__ctfe) return 0;
			return typeinfo.getHash(&value);
		}

		public this(){ this(T.init); }
		public this(T value){
			this.value = value;
			static if(needHash) hash = getHash;
		}
		override public const const(void)[] getVoidArr(){
			return [value];
		}
		protected override bool isTypeEq(Token t){ return t.isTypeOf!T(); }
		public override Token dup(){
			auto result = new TokenImpl!T(value);
			copy(result);
			return result;
		}
		protected override bool opEqualsToken(Token t){
			auto dest = cast(TokenImpl!T)t;
			return dest && value == dest.value;
		}
		public override int opCmp(Object o){
			auto dest = cast(typeof(this))o;
			if(!dest) return -1;
			if(this == dest) return 0;
			static if(is(typeof(value.opCmp(dest.value)))){
				return value.opCmp(dest.value);
			}else static if(is(typeof(value - dest.value) : int)){
				return value - dest.value;
			}else{
				static if(is(typeof(()=>value>dest.value))){
					return value == dest.value? 0: value > dest.value? 1: -1;
				}else{
					if(value == dest.value) return 0;
					assert(false, "not implemented");
				}
			}
		}
		public override hash_t toHash(){
			if(!isEmpty){
				static if(needHash){
					return hash;
				}else{
					return value.toHash();
				}
			}else{
				return typeinfo.toHash();
			}
		}
		protected override const string toStringType(){ return T.stringof; }
		protected override const string toStringValue(){
			static if(is(T : string)){
				return "\""~cast(string)value~"\"";
			}else static if(is(T : dchar)){
				return "'"~to!string(value)~"'";
			}else{
				return to!string(value);
			}
		} 
	}

	public static Token getToken(T)(size_t line, size_t num){
		auto t = new TokenImpl!T();
		t.line = line; t.num = num;
		return t;
	}
	public static Token getToken(T)(T value, size_t line, size_t num){
		auto impl = new TokenImpl!T(value);
		impl.isEmpty = false;
		impl.line = line; impl.num = num;
		return impl;
	}

	protected abstract bool isTypeEq(Token t){ return false; }
	protected abstract bool opEqualsToken(Token dest){ return false; }
	public override bool opEquals(Object o){
		if(this is o) return true;
		auto t = cast(Token) o;
		if(isEmpty) return t.isEmpty && this.isTypeEq(t);
		return !t.isEmpty && opEqualsToken(t);
	}
	protected abstract Token dup();
	protected void copy(Token t){ t.isEmpty = isEmpty; }
	public const bool isTypeOf(T)(){
		return cast(TokenImpl!T)this !is null;
	}
	public inout T peek(T)(){
		assert(!isEmpty, "Can't peek empty token");
		auto impl = cast(TokenImpl!T)this;
		assert(impl !is null, "Token is not type of "~T.stringof~" but "~toStringType());
		return impl.value;
	}
	public Token empty(){
		Token result = dup();
		result.isEmpty = true;
		return result;
	}
	abstract public const const(void[]) getVoidArr();
	abstract protected const string toStringType();
	abstract protected const string toStringValue();
	public override const string toString(){
		if(isEmpty) return toStringType();
		return toStringType()~"("~toStringValue()~")";
	}
}
Token token(T)(){ return token!T(0, 0); }
Token token(T)(size_t num, size_t line){
	return Token.getToken!T(num, line);
}
Token token(T)(T value){ return token(value, 0, 0); }
Token token(T)(T value, size_t num, size_t line){
	return Token.getToken(value, num, line);
}

struct var_id{
	union{
		ptrdiff_t id;
		void* ptr;
	}
	this(ptrdiff_t v){ id = v; }
	alias id this;
}
enum func_id : size_t { init=size_t.init }

alias void delegate(Context c) compilable;
interface Context{
	public final void defLocal(T)(string name){ defLocalImpl(TypeHandler.instance!T, name); }
	protected void defLocalImpl(TypeHandler t, string name);
	public var_id getLocalPos(string name);
	public compilable getLocal(string name);
	public compilable getFunc(string name);
	final public compilable getConstant(T)(T value){ return this.getConstantImpl(TypeHandler.instance!T, token(value)); }
	final public compilable getConstant(T : Token)(Token value){ return this.getConstantImpl(TypeHandler.instance!T, value); }
	protected compilable getConstantImpl(TypeHandler t, const Token val);
	public void defFunc(string name, compilable func);
	final public void addArg(T)(string name){ addArgImpl(TypeHandler.instance!T, name); }
	protected void addArgImpl(TypeHandler t, string name);
	public void callFunc(string name, const compilable[] args...);
	public void return_(compilable val);
	public void startScope();
	public void endScope();
	public void while_(compilable cond, compilable stmt);
	public void value(TypeHandler t, const void[] v);
	final public void convert(T1, T2)(){ convertImpl(T1.stringof, T2.stringof); }
	protected void convertImpl(string A, string B);
	public void callOpe(string name, const compilable[] tree...);
}

alias Syntax(alias startSymbol, SyntaxTreeType_, rules...) = Syntax!(true, startSymbol, SyntaxTreeType_, rules);

template Syntax(bool ctfeTable, alias startSymbol, SyntaxTreeType_, rules...){
	alias SyntaxTreeType_ SyntaxTreeType;
	enum nonterm[] nonterms = NontermList!rules;
	Object exec(size_t num, Object[] param){
		switch(num){
			foreach(i, e; ExpressionList!(rules)){
			case i:
				alias ActionParams!(e.symbols) paramType;
				paramType params;
				foreach(j, s; e.symbols){
					static if(is(typeof(s) == nonterm)) {
						params[j] = cast(SyntaxTreeType)param[j];
					}else{
						params[j] = cast(Token)param[j];
					}
				}
				return e.action(params);
			}
		default:
		}
		assert(false, "action "~num.to!string~" not found");
	}

	alias ExpressionList() = TypeTuple!();
	alias ExpressionList(rules...) = TypeTuple!(rules[0].rules, ExpressionList!(rules[1..$]));

	alias ActionParams() = TypeTuple!();
	alias ActionParams(symbols...) = TypeTuple!(Select!(is(typeof(symbols[0])==nonterm), SyntaxTreeType, Token), ActionParams!(symbols[1..$]));

	alias TerminalList!rules terminals;
//	enum terminalDic = createTerminalDictionary!(terminals.list)();
//	alias ActionList!rules actions;
	enum syntaxTable = syntaxTables!(terminals, nonterms, rules) ~ Production(Symbol.start, [symbol!nonterms(startSymbol)]);
	static if(ctfeTable){
		enum result = LaLrParser!(Symbol, SyntaxTreeType, nonterms, terminals, exec)(syntaxTable);
		static if(result.srConflictNum > 0) pragma(msg, "shift-reduce confliction:"~to!string(result.srConflictNum));
		pragma(msg, "table created "~result.actionTable.length.to!string);
	}else{
		immutable LaLrParser!(Symbol, SyntaxTreeType, nonterms, terminals, exec) result;
		static this(){
			result = LaLrParser!(Symbol, SyntaxTreeType, nonterms, terminals, exec)(syntaxTable);
			import std.stdio;
			if(result.srConflictNum > 0){
				write("shift-reduce confliction:",result.srConflictNum);
			}
		}
	}
	auto parse(T)(T tokenizer){ return result.parse(tokenizer); }
}

abstract class Tree{
	alias toCompilable this;
	final const compilable toCompilable(){ return &ast2ir; }
	final const T opCast(T)() if(is(T == class) || is(T == interface)){
		return cast(T)super;
	}
	abstract void ast2ir(Context);
}