module libdsl.compiler;
public import libdsl.common;
import libdsl.package_common;
import libdsl.interpreter;
import std.algorithm;
import std.traits;
import std.typetuple;
import std.typecons;
import std.conv;

var_id toVarid(void* ptr){
	var_id id;
	id.ptr = ptr;
	return id;
}

private auto curry(alias fun, args...)(args a){
	static class c{
		args a;
		this(args a){ this.a = a; }
		auto func(ParameterTypeTuple!fun[args.length..$] param){
			return fun(a, param);
		}
	}
	return &(new c(a).func);
}

private template staticIota(size_t start, size_t end, size_t step = 1) if (step > 0){
    static if (start+1 >= end){
        static if (start >= end){
            alias TypeTuple!() staticIota;
        }else{
            alias TypeTuple!(start) staticIota;
        }
    }else{
        enum mid = start + (end - start)/2;
        alias staticIota = TypeTuple!(staticIota!(start, mid), staticIota!(mid, end));
    }
}

template Compiler(funcs...){
	alias CompilerContext!funcs context;
	struct UnionTuple(types...){
		size_t i;
		union { private types fields; }
		const string toString(){
			final switch(i){
				foreach(i, T; types){
					mixin("case "~i.to!string~":");
					return fields[i].to!string;
				}
			}
			assert(false);
		}
		auto set(const Token token){
			foreach(i_, T; types){
				if(token.isTypeOf!T){
					i = i_;
					fields[i_] = token.peek!T;
					return;
				}
			}
			assert(false, token.toString~" can't set into "~types.stringof);
		}
		auto set(T)(T value){
			enum index = staticIndexOf!(T, types);
			i = index;
			return fields[index] = value;
		}
		const auto get(T)(){ return fields[staticIndexOf!(T, types)]; }
		const auto get(size_t i)(){ return fields[i]; }
	}
	class CompileContext : Context{
		Func func;
		var_id[string] names;
		var_id next;
		override void defLocalImpl(TypeHandler t, string name){
			func.code ~= ir!"defLocal"(t.typename, name);
		}
		override var_id getLocalPos(string name){
			return names[name];
		}
		override compilable getLocal(string name){
			static void compile(string name, Context context){
				auto c = cast(CompileContext)context;
				c.func.code ~= ir!"getLocal"(name);
			}
			return curry!compile(name);
		}
		override compilable getFunc(string name){
			static void compile(string name, Context context){
				auto c = cast(typeof(this))context;
				c.func.code ~= ir!"getFunc"(name);
			}
			return curry!compile(name);
		}
		override compilable getConstantImpl(TypeHandler t, const Token val){
			static void compile(TypeHandler t, const Token val, Context context){
				auto c = cast(typeof(this))context;
				c.func.code ~= ir!"getConstant"(t.typename, val);
			}
			return curry!compile(t, val);
		}
		override void defFunc(string name, compilable f){
			auto c = new typeof(this);
			f(c);
			func.funcs[name] = c.func;
		}
		override void addArgImpl(TypeHandler t, string name){
			func.args ~= tuple(t.typename, name, t.mangled);
		}
		override void callFunc(string name, const compilable[] args...){
			import std.conv;
			auto c = new typeof(this);
			const(Ir[])[] roots;
			foreach(i,a; args){
				c.func.code.length = 0;
				a(c);
				roots ~= c.func.code;
			}
			auto callFuncIr = ir!"callFunc"(name,roots);
			func.code ~= callFuncIr;
		}
		override void return_(compilable val){
			auto c = new typeof(this);
			val(c);
			func.code ~= ir!"return_"(c.func.code);
		}
		override void startScope(){
			func.code ~= ir!"startScope";
		}
		override void endScope(){
			func.code ~= ir!"endScope";
		}
		override void while_(compilable cond, compilable stmt){
			auto c = new typeof(this);
			cond(c);
			auto cRoot = c.func.code;
			c.func.code.length = 0;
			stmt(c);
			auto sRoot = c.func.code;
			func.code ~= ir!"while_"(cRoot, sRoot);
		}
		override void value(TypeHandler t, const void[] v){
			func.code ~= ir("value "~t.typename~"("~t.toStringValue(v.ptr)~")");
		}
		override void convertImpl(string A, string B){
			func.code ~= ir("convert "~A~"->"~B);
		}
		override void callOpe(string name, const compilable[] tree...){
			auto c = new typeof(this);
			Ir[][] roots;
			foreach(t; tree){
				c.func.code.length = 0;
				t(c);
				roots ~= c.func.code;
			}
			func.code ~= ir!"callOpe"(name, roots);
		}
	}
	struct getConstantData{
		string typename;
		UnionTuple!(context.usedTypes) value;
		this(string typename, const Token token){
			this.typename = typename;
			value.set = token;
		}
		const string toString(size_t level){ return indent(level)~"getConstant "~value.to!string~" : "~typename; }
	}
	struct callFuncData{
		string name;
		const Ir[][] args;
		const string toString(size_t level){
			string ret = indent(level)~"callFunc "~name;
			foreach(i,as; args){
				ret ~= "\n"~indent(level)~"arg "~i.to!string~":\n"~as.toString(level+2);
			}
			return ret;
		}
	}
	struct callOpeData{
		string name;
		const Ir[][] args;
		const string toString(size_t level){
			string ret = indent(level)~"callOpe "~name;
			foreach(i, a; args){
				ret ~= "\n"~indent(level)~"arg "~i.to!string~":\n"~a.toString(level+1);
			}
			return ret;
		}
	}
	struct while_Data{
		const Ir[] cond;
		const Ir[] stmt;
		const string toString(size_t level){
			return indent(level)~"while\n"~indent(level+1)~"cond:\n"~cond.toString(level+2)~"\n"~indent(level+1)~"stmt:\n"~stmt.toString(level+2);
		}
	}
	struct return_Data{
		const Ir[] code;
		const string toString(size_t level){ return indent(level)~"return\n"~code.toString(level+1); }
	}
	private struct Func{
		Tuple!(string, string, string)[] args;
		Func[string] funcs;
		Ir[] code;
		const string toString(size_t level = 0){
			string ret;
			if(code.length == 0) return "";
			foreach(i, a; args){
				ret ~= "\n"~indent(level)~"arg"~i.to!string~" : "~a[0]~" "~a[1];
			}
			foreach(name, f; funcs){
				ret ~= "\n"~indent(level)~"func "~name~"\n"~f.toString(level+1);
			}
			foreach(c; code){
				ret ~= "\n";
				ret ~= c.toString(level);
			}
			return ret[1..$];
		}
	}
	private struct Ir{
		IrType type;
		union {
			defLocalData[] defLocal;
			getLocalData[] getLocal;
			getFuncData[] getFunc;
			getConstantData[] getConstant;
			callFuncData[] callFunc;
			return_Data[] return_;
			startScopeData[] startScope;
			endScopeData[] endScope;
			while_Data[] while_;
			valueData[] value;
			convertData[] convert;
			callOpeData[] callOpe;
		}
		const string toString(){ return toString(0); }
		const string toString(size_t level){
			import std.conv;
			final switch(type){
			case IrType.defLocal: return defLocal[0].toString(level);
			case IrType.getLocal: return getLocal[0].toString(level);
			case IrType.getFunc: return getFunc[0].toString(level);
			case IrType.getConstant: return getConstant[0].toString(level);
			case IrType.callFunc: return callFunc[0].toString(level);
			case IrType.return_: return return_[0].toString(level);
			case IrType.startScope: return startScope[0].toString(level);
			case IrType.endScope: return endScope[0].toString(level);
			case IrType.while_: return while_[0].toString(level);
			case IrType.value: return value[0].toString(level);
			case IrType.convert: return convert[0].toString(level);
			case IrType.callOpe: return callOpe[0].toString(level);
			}
		}
	}
	string toString(const Ir[] irs, size_t level){
		string ret;
		foreach(a; irs){
			ret ~= "\n";
			ret ~= a.toString(level);
		}
		return ret.length? ret[1..$]: "";
	}
	auto ir(string name, T...)(T args){
		Ir ir;
		ir.type = mixin("IrType."~name);
		mixin("ir."~name~"=[typeof(ir."~name~"[0])(args)];");
		return ir;
	}
	auto ir(string name){
		return Ir.init;
	}
	auto ir(IrType type){
		Ir ret;
		ret.type = type;
		return ret;
	}
	struct Compiler(alias tree){
		pragma(msg, compile.toString);
		static auto compile(){
			auto context = new CompileContext;
			tree()(context);
			return context.func;
		}
		auto opCall(T...)(T args){ return exec(args); }
		R execFor(R, T...)(T args){ return exec(args); }
		auto exec(T...)(T arg){
			return implFunc!(context, compile).exec(arg);
		}
	}
}
private template CompilerContext(funcs_...){
	alias funcs_ funcs;
	enum opeParams = makeOpeParams!funcs;
	alias UsedTypes!funcs usedTypes;
	enum typeMap = makeTypeMap;
	auto makeTypeMap(){
		size_t[string] result;
		foreach(i, T; usedTypes){
			pragma(msg, "T:"~T.stringof~"->"~T.mangleof);
			result[T.mangleof] = i;
		}
		return result;
	}
	auto makeConvertible(types...)(){
		bool[Tuple!(string,string)] ret;
		foreach(i,t1; types){
			foreach(j,t2; types){
				static if(i != j){
					static if(is(t1 : t2)){
						ret[tuple(t1.mangleof, t2.mangleof)] = true;
					}
				}
			}
		}
		return ret;
	}
	enum convertible = makeConvertible!usedTypes;
	public bool isConvertible(in TypeHandler a, in TypeHandler b){
		if(a.mangled == b.mangled || tuple(a.mangled,b.mangled) in convertible) return true;
		if(a.mangled == func_id.mangleof && b.callable) return true;
		return false;
	}
}

private string makeArgType(alias context)(const Tuple!(string,string,string)[] args){
	string ret;
	foreach(a; args){
		import std.conv;
		assert(a[2] in context.typeMap, a[0]~" is not found in "~context.typeMap.to!string~" ("~a[2]~")");
		ret ~= "context.usedTypes["~context.typeMap[a[2]].to!string~"],";
	}
	return "TypeTuple!("~ret~")";
}
private template implFunc(alias context, alias func){
	mixin("alias "~makeArgType!context(func.args)~" AT;");
	static string makeVariableStruct(){
		string ret = "struct Vars{";
		foreach(i; 0..AT.length){
			ret ~= "AT["~i.to!string~"] "~func.args[i][1]~";";
		}
		ret ~= "ref auto get(string name)(){return mixin(name);}";
		foreach(name, f; func.funcs){
			ret ~= "alias implFunc!(context, func.funcs[\""~name~"\"]) "~name~";";
		}
		ret ~= "template getFunc(string name){mixin(\"alias \"~name~\" getFunc;\");}";
		return ret~"}";
	}
	mixin(makeVariableStruct());
	auto exec(AT args){
		Vars vars;
		alias OperatorContext!(context, func, vars, "code") oc;
			
		foreach(i, t; AT){
			static assert(t.mangleof==func.args[i][2] || mixin("is("~t.stringof~":"~func.args[i][0]~")"), "argument type mismatched "~t.stringof~","~func.args[i][0]);
			mixin("vars."~func.args[i][1]~"=args["~i.to!string~"];");
		}
		mixin(makeCodeImpl(func.code, 0, "oc"));
	}
}

private template OperatorContext(alias compilerContext, alias func_, alias env_, string irs_){
	alias compilerContext context;
	alias func_ func;
	mixin("enum irs=func."~irs_~";");
	alias env_ env;
	template ir(size_t i){
		mixin("enum ir=func."~irs_~"["~i.to!string~"];");
	}
	
	template next(size_t i, string code){
		alias OperatorContext!(context, func_, env_, irs_~"["~i.to!string~"]."~code) next;
	}
}

private string makeCodeImpl(T)(const T[] code, size_t num, string name){
	string ret;
	foreach(i,c; code){
		string args = name~","~(i+num).to!string;
		ret~="/*"~c.type.to!string~"*/";
		final switch(c.type){
		case IrType.defLocal: break;
		case IrType.getLocal:
			ret ~= "execGetLocal!("~args~");";
			break;
		case IrType.getFunc:
			ret ~= "&execGetFunc!("~args~");";
			break;
		case IrType.getConstant:
			ret ~= "execGetConstant!("~args~");";
			break;
		case IrType.callFunc: break;
		case IrType.return_:{
			ret ~= "return execReturn!("~args~");";
			break;
		}
		case IrType.startScope: break;
		case IrType.endScope: break;
		case IrType.while_: break;
		case IrType.value: break;
		case IrType.convert: break;
		case IrType.callOpe:{
			ret ~= "execCallOpe!("~args~");";
			break;
		}
		}
	}
	return ret;
}

private template execGetLocal(alias context, size_t i){
	enum ir = context.ir!i;
	static assert(ir.type == IrType.getLocal);
	enum name = ir.getLocal[0].name;
	auto execGetLocal(){
		return toVarid(&context.env.get!(name)());
	}
}
//TODO: 関数がネストした時オーバーロードするかどうかは怪しいので調べる
private template execGetFunc(alias context, size_t i){
	enum ir = context.ir!i;
	static assert(ir.type == IrType.getFunc);
	enum name = ir.getFunc[0].name;
	alias context.env.getFunc!name.exec execGetFunc;
}
private template execGetConstant(alias context, size_t i){
	enum ir = context.ir!i;
	static assert(ir.type == IrType.getConstant);
	enum valueType = ir.getConstant[0].value.i;
	enum execGetConstant = ir.getConstant[0].value.get!valueType;
}
private template execReturn(alias context, size_t i){
	enum ir = context.ir!i;
	static assert(ir.type == IrType.return_);
	alias context.next!(i, "return_[0].code") next;
	auto execReturn(){
		static if(next.irs.length > 1){
			mixin(makeCodeImpl(next.irs[0..$-1], 0, "next"));
		}
		mixin("return "~makeCodeImpl(next.irs[$-1..$], next.irs.length-1, "next"));
	}
}
private template toType(alias v){
	alias Unqual!(typeof(v)) toType;
}
private auto toTypeHandlers(T...)(){
	TypeHandler[] ret;
	foreach(t; T) ret ~= TypeHandler.instance!t;
	return ret;
}
private template execCallOpe(alias context, size_t i){
	enum ir = context.ir!i;
	static assert(ir.type == IrType.callOpe);
	enum data = ir.callOpe[0];
	enum name = data.name;
	string makeArgTupleSrc(size_t num){
		string ret = "TypeTuple!(";
		foreach(i; 0..num) ret ~= "arg"~i.to!string~",";
		return ret~")";
	}
	size_t findOpe(in OpeParam[] opes, TypeHandler[] argTypes){
		auto result = findOverloads!(context.context.isConvertible)(opes, argTypes);
		assert(result.length > 0, "operator "~name~"("~argTypes.to!string~") is not mached\n"~opes.to!string);
		assert(result.length == 1, "operator "~name~"("~argTypes.to!string~") is mached definitions\n"~opes.to!string);
		return result[0].id;
	}
	string makeArgsSrc(size_t num){
		string ret;
		foreach(i; 0..num){
			ret ~=
				"alias context.next!(i, \"callOpe[0].args["~i.to!string~"]\") next"~i.to!string~";"
				"static if(next"~i.to!string~".irs.length > 1){"
				"	mixin(makeCodeImpl(next"~i.to!string~".irs[0..$-1], 0, \"next"~i.to!string~"\"));"
				"}"
				"mixin(\"auto arg"~i.to!string~"=\"~makeCodeImpl(next"~i.to!string~".irs[$-1..$], next"~i.to!string~".irs.length-1, \"next"~i.to!string~"\"));";
		}
		return ret;
	}
	string makeBranchesSrc(size_t offset, size_t num, string tupleName){
		string ret, tuple;
		tuple = "alias TypeTuple!(";
		foreach(i; offset..offset+num){
			ret ~=
				"alias context.next!(i, \"callOpe[0].args["~i.to!string~"]\") next"~i.to!string~";"
				"auto branch"~i.to!string~"(){"
				"	mixin(makeCodeImpl(next"~i.to!string~".irs[0..$-1], 0, \"next"~i.to!string~"\"));"
				"	mixin(\"result=\"~makeCodeImpl(next"~i.to!string~".irs[$-1..$], next"~i.to!string~".irs.length-1, \"next"~i.to!string~"\"));"
				"}";
			tuple ~= "branch"~i.to!string~",";
		}
		return ret~tuple~") "~tupleName~";";
	}
	auto execCallOpe(){
		enum params = context.context.opeParams[data.name];
		assert(params.length);
		mixin(makeArgsSrc(params[0].argTypes.length));
		mixin("alias "~makeArgTupleSrc(params[0].argTypes.length)~" argsTuple;");
		alias context.context.funcs[findOpe(params, toTypeHandlers!(staticMap!(toType, argsTuple)))] ope;
		ope.RetType result;
		mixin(makeBranchesSrc(params[0].argTypes.length, params[0].branches, "branchesTuple"));
		template RuntimeContext(){
			ref T getLocalVar(T)(var_id id){
				return *cast(T*)id.ptr;
			}
			void setResult(ope.RetType val){
				result = val;
			}
			template call(T){
				T call(A...)(A args){
					return args[0](args[1..$]);
				}
			}
			alias ope.option option;
		}
		ope.exec!(RuntimeContext!(), argsTuple, branchesTuple);
		return result;
	}
}

enum IrType {defLocal, getLocal, getFunc, getConstant, callFunc, return_, startScope, endScope, while_, value, convert, callOpe}
struct defLocalData{
	string typename;
	string name;
	const string toString(size_t level){ return level.indent~"defLocal "~name~" : "~typename; }
}
struct getLocalData{
	string name;
	const string toString(size_t level){ return level.indent~"getLocal "~name; }
}
struct getFuncData{
	string name;
	const string toString(size_t level){ return indent(level)~"getFunc "~name; }
}
struct startScopeData{
	const string toString(size_t level){ return indent(level)~"startScope"; }
}
struct endScopeData{
	const string toString(size_t level){ return indent(level)~"endScope"; }
}
struct valueData{
	const string toString(size_t level){ return "value"; }
}
struct convertData{
	const string toString(size_t level){ return "convert"; }
}

string indent(size_t level){
	return "                                                 "[0..level*4];
}